/* Zaurus framebuffer VNC server

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version
   2 of the License, or (at your option) any later version.

   (c) Pierre-Philippe Coupard, 27/01/2002
*/
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <signal.h>
#include "rfb.h"
#include "keysym.h"
#include "keysym2scancode.h"
#include "loadjpeg.h"
#include "panelzones.h"
#include "leds.h"



#ifdef IPAQ
#define FANCY_PANEL_JPG "/usr/local/vnc/ipaq_panel.jpg"
#define DESKTOP_NAME "iPAQ"
#else   /* #ifdef ZAURUS */
#define FANCY_PANEL_JPG "/home/root/Applications/fbvncserver/zaurus_panel.jpg"
#define DESKTOP_NAME "Zaurus"
#endif

#define DOUBLE_KEYS_DELAY       750     /* milliseconds */
#define LED_ORANGE_COLOR (0x1f|(0x11<<5)|(0x08<<10))    /* Pale-orange color */
#define LED_GREEN_COLOR  (0x1a|(0x1f<<5)|(0x0a<<10))    /* Pale-green color */

/* Types */
struct ClientData
{
  Bool oldButton;
  int oldx,oldy;
};

typedef struct
{
  long x;
  long y;
  long pressure;
} ARTIFICIAL_TS_EVENT;

struct ledlight
{
  unsigned short int pixels[50][2];     /* Max. 50 pixels */
  int min_x,min_y,max_x,max_y;          /* Dimensions of a box containing all
                                           the pixels */
  int prev_state;
};



/* Variables */
static int fancy_display=0;
static int split_display=0;
static int got_event=0;                /* We need to process events in priority */
static int last_pen_down=0;
static int double_pixel = 0;
static int rotate_cw = 0;
static int noleds;
static unsigned short int display_x_offset = DISPLAY_X_OFFSET;
static unsigned short int display_y_offset = DISPLAY_Y_OFFSET;
struct fb_var_screeninfo vscrinfo;
#ifndef IPAQ
struct ledlight led[2]={{LED0_PIXELS,
                        LED0_MIN_X,LED0_MIN_Y,LED0_MAX_X,LED0_MAX_Y,
                        -1},
                        {LED1_PIXELS,
                        LED1_MIN_X,LED1_MIN_Y,LED1_MAX_X,LED1_MAX_Y,
                        -1}};
#endif
int softkbdmap[][6]=SOFT_KBD_MAP;
int panel_bottom_x_offset=0;
int panel_bottom_y_offset=0;
int last_softkeydown_key=-1;
struct timeval last_softkeydown_time;

/* We put the following vars here so __asm__ is happy */
unsigned short int *fbbuf,*cmpfbbuf,*rfbbuf;
struct varblock_t       /* This horror is to simplify mem access in __asm__ */
{
  int min_i;
  int min_j;
  int max_i;
  int max_j;
  int r_offset;
  int g_offset;
  int b_offset;
  int rfb_xres;
  int rfb_maxy;
} varblock;
/* The volatile below is a hack to get the compiler to not optimize out the
* code left  after the assembly. */
volatile struct varblock_t *varblock_volatile = &varblock;
int rfb_width, rfb_height;


/* Emit a keyboard scancode sequence with key-ups or a key-downs. The sequence
   is terminated with 0xff */
int emit_scancode_sequence(unsigned char *s)
{
  int fd;
  int ret;

  ret = 0;
  if((fd=open("/dev/kbdsim",O_WRONLY))>=0)
  {
    while(*s!=0xff)
    {
      write(fd,s,2);
      s+=2;
    }
    close(fd);
  }
  else
  {
    ret=-1;
  }

  return(ret);
}



/* Emit a touchscreen event sequence, The sequence is terminated with an event
   that has x==-1 */
int emit_touchscreen_sequence(ARTIFICIAL_TS_EVENT *s)
{
  int fd;
  int ret;

  ret = 0;
  if((fd=open("/dev/tssim",O_WRONLY))>=0)
  {
    while(s->x!=-1)
    {
      write(fd,s,sizeof(ARTIFICIAL_TS_EVENT));
      s++;
    }
    close(fd);
  }
  else
    ret=-1;

  return(ret);
}



/* Pointer events handler */
void ptrevent(int buttonMask,int x,int y,rfbClientPtr cl)
{
  struct timeval time;
  long time_diff_ms;
  struct timezone ignored_tz;
  unsigned char scancode_sequence[5];
  ARTIFICIAL_TS_EVENT touchscreen_sequence[5];
  int i;

  if(rotate_cw) {
      int tmp;
      tmp = x;
      x = y;
      y = rfb_width - tmp;
  }
  if(double_pixel) {
    x /= 2;
    y /= 2;
  }

  got_event=1;

  /* Did the event occur inside the display proper, and no key-down is
     pending ? */
  if(x>=display_x_offset && x<display_x_offset+vscrinfo.yres &&
        y>=display_y_offset && y<display_y_offset+vscrinfo.xres &&
        last_softkeydown_key==-1)
  {
    x-=display_x_offset;
    y-=display_y_offset;

    if(buttonMask!=0)
    {
      for(i=0;i<(last_pen_down?1:5);i++)
      {
        touchscreen_sequence[i].x=x;
        touchscreen_sequence[i].y=y;
        touchscreen_sequence[i].pressure=1;
      }
      touchscreen_sequence[i].x=-1;
      emit_touchscreen_sequence(touchscreen_sequence);
    }
    else
    {
      if(last_pen_down)
      {
        touchscreen_sequence[0].x=x;
        touchscreen_sequence[0].y=y;
        touchscreen_sequence[0].pressure=0;
        touchscreen_sequence[1].x=-1;
        emit_touchscreen_sequence(touchscreen_sequence);
      }
    }
    last_pen_down=(buttonMask!=0);
  }
  else          /* The event occurred outside the display proper */
  {
    /* Do soft keyboard events only if we do the fancy display */
    if(fancy_display)
    {
      /* Is this a change of the state of the button ? */
      if(((last_pen_down!=0 && buttonMask==0) ||
                (last_pen_down==0 && buttonMask!=0)))
      {
        /* Has the user has clicked on a defined key, or do we have a key
           down pending ? */
        if(last_softkeydown_key!=-1)
          i=last_softkeydown_key;
        else
          for(i=0;softkbdmap[i][0]>-1;i++)
            if(x>=softkbdmap[i][0] && x<softkbdmap[i][0]+softkbdmap[i][2] &&
                y>=softkbdmap[i][1] && y<softkbdmap[i][1]+softkbdmap[i][3])
              break;
        if(softkbdmap[i][0]>-1)
        {
          /* Does the key have a secondary scancode that's different from
             its primary scancode ? */
          if(softkbdmap[i][4]!=softkbdmap[i][5])
          {
            /* If we have a keydown, we don't know what scancode to send yet*/
            if(buttonMask!=0)
            {
              gettimeofday(&last_softkeydown_time,&ignored_tz);
              last_softkeydown_key=i;
            }
            else        /* We have a keyup : send either the primary or */
            {           /* secondary scancode key down and key up */
              if(last_softkeydown_key!=-1)
              {
                gettimeofday(&time,&ignored_tz);
                time_diff_ms=(time.tv_sec-last_softkeydown_time.tv_sec)*1000+
                        (time.tv_usec-last_softkeydown_time.tv_usec)/1000;
                if(time_diff_ms>DOUBLE_KEYS_DELAY)
                {
                  scancode_sequence[0]=softkbdmap[last_softkeydown_key][5];
                  scancode_sequence[2]=softkbdmap[last_softkeydown_key][5];
                }
                else
                {
                  scancode_sequence[0]=softkbdmap[last_softkeydown_key][4];
                  scancode_sequence[2]=softkbdmap[last_softkeydown_key][4];
                }
                scancode_sequence[1]=1;
                scancode_sequence[3]=0;
                scancode_sequence[4]=0xff;
                emit_scancode_sequence(scancode_sequence);
                last_softkeydown_key=-1;
              }
            }
          }
          else  /* The key is a normal key without a special */
          {     /* secondary scancode */
            scancode_sequence[1]=(buttonMask!=0);
            scancode_sequence[2]=0xff;
            if(buttonMask!=0)
            {
              scancode_sequence[0]=softkbdmap[i][4];
              emit_scancode_sequence(scancode_sequence);
              last_softkeydown_key=i;
            }
            else
            {
              if(last_softkeydown_key!=-1)
              {
                scancode_sequence[0]=softkbdmap[last_softkeydown_key][4];
                emit_scancode_sequence(scancode_sequence);
                last_softkeydown_key=-1;
              }
            }
          }
        }

        last_pen_down=(buttonMask!=0);
      }
    }
  }

  return;
}



/* Keyboard events handler */
void keyevent(Bool down,KeySym key,rfbClientPtr cl)
{
  unsigned char *scancodes;

  got_event=1;

  scancodes = keysym2scancodes(down, key);
  if(down && (scancodes != NULL))
    emit_scancode_sequence(scancodes);

  return;
}



/* Client disconnection */
void clientleft(rfbClientPtr cl)
{
  free(cl->clientData);
}



/* Client connection */
static enum rfbNewClientAction newclient(rfbClientPtr cl)
{
  cl->clientData = (void*)calloc(sizeof(struct ClientData),1);
  cl->clientGoneHook = clientleft;
  return RFB_CLIENT_ACCEPT;
}



/* SIGPIPE signal handler : we catch SIGPIPE and ignore it so the library's
   write calls receive -EPIPE but the server keeps on running */
void sighandler(int signum)
{
  return;
}


/* Main program */
int main(int argc,char **argv)
{
  int fbfd;
  FILE *jpgfd;
  rfbScreenInfoPtr rfbScreen;
  short int i,j;
  unsigned short int *jpgbuf;
#ifndef IPAQ
  int led_states;
#endif

  fprintf(stderr, "FBVNCserver Version: %s\n",
#include "version.h"
  );

  jpgbuf = 0;
  /* See if we have arguments that we know of */
  for(i=1;i<argc;i++)
  {
    if(!strcmp(argv[i],"-h") || !strcmp(argv[i],"--help"))
    {
      printf("\n");
      printf("Usage: %s\n",argv[0]);
      printf("         [-fancy] [-split] [-double] [-cw]\n");
      printf("         [VNC server options]\n");
      printf("  Note that VNC Server Options must go last\n");
      printf("\n");
      printf("  -fancy  : display a fancy\n");
      printf("            Zaurus panel\n");
      printf("\n");
#ifndef IPAQ
      printf("  -split  : split the panel in two (implies -fancy)\n");
#endif
      printf("  -double : double the display in\n");
      printf("            each dimension. Excludes\n");
      printf("            fancy/split.\n");
      printf("  -noleds : don't simulate LEDs on the client\n");
      printf("            (Use for 5600)\n");
      printf("\n");
      printf("\n");
      return(-1);
    }
    else if(!strcmp(argv[i],"-noleds")) {
      noleds = 1;
    }
    else if(!strcmp(argv[i],"-fancy")) {
      fancy_display=1;
      double_pixel = 0;
      rotate_cw = 0;
    }
#ifndef IPAQ
    else if(!strcmp(argv[i],"-split")) {
      split_display=1;
      fancy_display=1;
      double_pixel = 0;
      rotate_cw = 0;
    }
#endif
    else if(!strcmp(argv[i],"-double")) {
      double_pixel = 1;
      fancy_display = 0;
      split_display = 0;
    }
    else if(!strcmp(argv[i],"-cw")) {
      fancy_display = 0;
      split_display = 0;
      rotate_cw = 1;
    } else {
        break;
    }
  }
  argc -= i - 1;
  argv += i - 1;

  /* Open the framebuffer */
  if((fbfd=open("/dev/fb0",O_RDONLY))<0)
  {
    perror("Opening /dev/fb0");
    return(-1);
  }

  /* Request the framebuffer's screen info */
  if(ioctl(fbfd,FBIOGET_VSCREENINFO,&vscrinfo)<0)
  {
    close(fbfd);
    perror("ioctl(FBIOGET_VSCREENINFO)");
    return(-1);
  }

  /* Allocate the comparison framebuffer and black it out */
  if((cmpfbbuf=(unsigned short int *)malloc(vscrinfo.xres * vscrinfo.yres * 2))
        ==NULL)
  {
    close(fbfd);
    perror("malloc");
    return(-1);
  }
  memset(cmpfbbuf,0,vscrinfo.xres*vscrinfo.yres*2);

  /* If we do fancy display, load the panel image */
  if(fancy_display)
  {
    if((jpgfd=fopen(FANCY_PANEL_JPG,"rb"))==NULL)
      fancy_display=0;
    else
    {
      if((jpgbuf=loadjpeg(jpgfd, &rfb_width, &rfb_height))==NULL)
      {
        fclose(jpgfd);
        fancy_display=0;
      }
      else
      {
        fclose(jpgfd);

        /* The panel must be bigger than the display and the offset, and if
           we do split screen, the height must be able to contain the display */
        if(rfb_width <= vscrinfo.yres+display_x_offset || rfb_height<=
                (vscrinfo.xres+display_y_offset)*(split_display?2:1))
        {
          free(jpgbuf);
          fancy_display=0;
          split_display=0;
        }
        else
        {
          /* Black out the display mask */
          for(j=display_y_offset;j<display_y_offset+vscrinfo.xres;j++)
            memset(jpgbuf+j*rfb_width+display_x_offset,0,vscrinfo.yres*2);
        }
      }
    }
  }
  if(!fancy_display)
  {
    split_display=0;
    rfb_width=vscrinfo.yres;
    rfb_height=vscrinfo.xres;
    display_x_offset=0;
    display_y_offset=0;
  }
  if(split_display)
  {
    rfb_height/=2;

#ifndef IPAQ
    /* Fixup the LEDs pixels coordinates and update zones */
    for(j=0;j<2;j++)
    {
      for(i=0;led[j].pixels[i][0];i++)
      {
        led[j].pixels[i][0]+=rfb_width;
        led[j].pixels[i][1]-=rfb_height;
      }
      led[j].min_x+=rfb_width;
      led[j].max_x+=rfb_width;
      led[j].min_y-=rfb_height;
      led[j].max_y-=rfb_height;
    }
#endif

    /* Fixup the soft keyboard's key coordinates */
    for(i=0;softkbdmap[i][0]>-1;i++)
    {
      softkbdmap[i][0]+=rfb_width;
      softkbdmap[i][1]-=rfb_height;
    }

    rfb_width*=2;
  }
  if(double_pixel){
    rfb_height *= 2;
    rfb_width  *= 2;
  }
  if(rotate_cw) {
      int tmp;
      tmp = rfb_height;
      rfb_height = rfb_width;
      rfb_width = tmp;
  }

  /* Allocate and initialize the remote framebuffer */
  if((rfbbuf=(unsigned short int *)malloc(rfb_width*rfb_height*
                                        vscrinfo.bits_per_pixel/8))==NULL)
  {
    close(fbfd);
    free(cmpfbbuf);
    perror("malloc");
    return(-1);
  }

  if(fancy_display)
  {
    if(!split_display)
      memcpy(rfbbuf,jpgbuf,(rfb_width*rfb_height*vscrinfo.bits_per_pixel)/8);
    else
    {
      for(i=0;i<rfb_height;i++)
      {
        memcpy(rfbbuf+i*rfb_width,
                jpgbuf+(i*rfb_width)/2,
                ((rfb_width/2)*vscrinfo.bits_per_pixel)/8);
        memcpy(rfbbuf+i*rfb_width+(rfb_width/2),
                jpgbuf+((i+rfb_height)*rfb_width)/2,
                ((rfb_width/2)*vscrinfo.bits_per_pixel)/8);
      }
    }
    free(jpgbuf);
  }

  /* mmap the fb */
  if((fbbuf=(unsigned short int *)mmap(0,vscrinfo.xres*vscrinfo.yres*
                                        vscrinfo.bits_per_pixel/8,
                                        PROT_READ,MAP_SHARED,fbfd,
                                        vscrinfo.yoffset*vscrinfo.xres*
                                        vscrinfo.bits_per_pixel/8))<0)
  {
    close(fbfd);
    free(cmpfbbuf);
    free(rfbbuf);
    perror("mmap");
    return(-1);
  }

  /* Initialize the rfb server */
  rfbScreen                  = rfbGetScreen(&argc,argv,rfb_width,rfb_height,5,2,2);
  rfbScreen->desktopName     = DESKTOP_NAME;
  rfbScreen->frameBuffer     = (char *)rfbbuf;
  rfbScreen->rfbAlwaysShared = TRUE;
  rfbScreen->ptrAddEvent     = ptrevent;
  rfbScreen->kbdAddEvent     = keyevent;
  rfbScreen->newClientHook   = newclient;
  rfbScreen->httpDir         = "/home/root/Applications/fbvncserver/classes";
  rfbInitServer(rfbScreen);

  /* Make sure we catch EPIPE so we don't get killed by closed sockets when
     the library tries to write to them */
  signal(SIGPIPE,sighandler);

  /* Mark the whole screen as modified the first time */
  rfbMarkRectAsModified(rfbScreen,0,0,rfb_width,rfb_height);

  /* Do our own rfb event loop, and also look for changes in the fb and update
     the rfb accordingly */
  varblock.r_offset = vscrinfo.red.offset+vscrinfo.red.length-5;
  varblock.g_offset = vscrinfo.green.offset+vscrinfo.green.length-5;
  varblock.b_offset = vscrinfo.blue.offset+vscrinfo.blue.length-5;
  varblock.rfb_xres = vscrinfo.yres;
  varblock.rfb_maxy = vscrinfo.xres-1;
  while(1)
  {
      /* If no clients are connected, just wait for events and don't bother
         updating the rfb, to save CPU */
      while(rfbScreen->rfbClientHead==NULL) {
          rfbProcessEvents(rfbScreen,100000);       /* We still run a tight loop
                                                       so http clients don't wait */
      }
      /* Process events for a while */
      do
      {
          got_event=0;
          rfbProcessEvents(rfbScreen,30000);
      }
      while(got_event);

      varblock.min_i = varblock.min_j = 9999;
      varblock.max_i = varblock.max_j = -1;

      if(rotate_cw && double_pixel) {
          unsigned int *f, *c, *r;
          int x, y;

          f = (unsigned int *)fbbuf;
          c = (unsigned int *)cmpfbbuf;
          r = (unsigned int *)rfbbuf;
          for(y = 0; y < vscrinfo.yres; y++) {
              for(x = 0; x < vscrinfo.xres; x += 2) {
                  if(*f != *c) {
                      unsigned int high, low;
                      *c = *f;
                      /* Here we are exracting the high and low pixels.  (f
                       * points at a pair of pixels.)  We first extract each
                       * pixel, then we duplicate it into the adjacent 16
                       * bits. */
                      high  = *f & 0xffff0000;
                      low   = *f & 0x0000ffff;
                      low  |= low  << 16;
                      high |= high >> 16;

                      /* Do color mods */
                      low =  ((low >> varblock.r_offset) & 0x1f001f)        |
                            (((low >> varblock.g_offset) & 0x1f001f) <<  5) |
                            (((low >> varblock.b_offset) & 0x1f001f) << 10);

                      high =  ((high >> varblock.r_offset) & 0x1f001f)        |
                              (((high >> varblock.g_offset) & 0x1f001f) <<  5) |
                              (((high >> varblock.b_offset) & 0x1f001f) << 10);

                      /*
                       * Store it in (x,y), (x+1,y), (x,y+1), (x+1, y+1)
                       * because we're make each pixel four pixels.
                       */
                      *r = low;
                      *(r + rfb_width/2) = low;
                      r++;
                      *r = high;
                      *(r + rfb_width/2) = high;
                      r++;

                      /* Mark the boundaries if the changes. */
                      if(x < varblock.min_i) varblock.min_i = x;
                      else if(x > varblock.max_i) varblock.max_i = x;
                      if(y > varblock.max_j) varblock.max_j = y;
                      else if(y < varblock.min_j) varblock.min_j = y;
                  } else {
                      r += 2;
                  }
                  f++, c++;
              }
              r += rfb_width/2; /* Skip past second row of pixels. They're written already. */
          }
          if(varblock.min_i < 9999) {
              /* Update the remote buffer */
              if(varblock.max_i < 0) varblock.max_i = varblock.min_i;
              if(varblock.max_j < 0) varblock.max_j = varblock.min_j;
              rfbMarkRectAsModified(rfbScreen, ((varblock.min_i - 2) * 2) + display_x_offset,
                  ((varblock.min_j - 2) * 2) + display_y_offset, ((varblock.max_i + 2) * 2 + 1) +
                  display_x_offset + 1, ((varblock.max_j + 2) * 2) + display_y_offset + 1);
              rfbProcessEvents(rfbScreen, 10000);
          }
      } else if(rotate_cw) {
          unsigned int *f, *c, *r;
          int x, y;

          f = (unsigned int *)fbbuf;
          c = (unsigned int *)cmpfbbuf;
          r = (unsigned int *)rfbbuf;
          for(y = 0; y < vscrinfo.yres; y++) {
              /* for(x = vscrinfo.xres - 1; x >= 0; x -= 2) { */
              for(x = 0; x < vscrinfo.xres; x += 2) {
                  if(*f != *c) {
                      unsigned int pixel;
                      *c = *f;
                      pixel = *f;

                      pixel =  ((pixel >> varblock.r_offset) & 0x1f001f)        |
                              (((pixel >> varblock.g_offset) & 0x1f001f) <<  5) |
                              (((pixel >> varblock.b_offset) & 0x1f001f) << 10);

                      *r = pixel;

                      if(x < varblock.min_i) varblock.min_i = x;
                      else if(x > varblock.max_i) varblock.max_i = x;
                      if(y > varblock.max_j) varblock.max_j = y;
                      else if(y < varblock.min_j) varblock.min_j = y;
                  }
                  f++, c++;
                  r++;
              }
          }
          if(varblock.min_i < 9999) {
              if(varblock.max_i < 0) varblock.max_i = varblock.min_i;
              if(varblock.max_j < 0) varblock.max_j = varblock.min_j;
              rfbMarkRectAsModified(rfbScreen, varblock.min_i + display_x_offset,
                  varblock.min_j + display_y_offset, varblock.max_i +
                  display_x_offset + 2, varblock.max_j + display_y_offset + 1);
              rfbProcessEvents(rfbScreen, 10000);
          }
      } else if(double_pixel) {
          unsigned int *f, *c, *r_column, *r;
          int x, y;

          f = (unsigned int *)fbbuf;
          c = (unsigned int *)cmpfbbuf;
          /* This next line could be pre-computed to save a few cycles */
          r_column = (unsigned int *)(rfbbuf + (vscrinfo.yres * 2) * ((vscrinfo.xres - 1) * 2));
          for(y = 0; y < vscrinfo.yres; y++) {
              r = r_column;
              for(x = vscrinfo.xres - 1; x >= 0; x -= 2) {
                  if(*f != *c) {
                      unsigned int high, low;
                      *c = *f;
                      low = (*f) & 0xffff;
                      high = *f & 0xffff0000;

                      low =  ((low >> varblock.r_offset) & 0x1f)        |
                          (((low >> varblock.g_offset) & 0x1f) <<  5) |
                          (((low >> varblock.b_offset) & 0x1f) << 10);
                      low |= low << 16;

                      high = ((high >> varblock.r_offset) & 0x1f0000)        |
                          (((high >> varblock.g_offset) & 0x1f0000) <<  5) |
                          (((high >> varblock.b_offset) & 0x1f0000) << 10);
                      high |= high >> 16;

                      *r = low;
                      r -= vscrinfo.yres;
                      *r = low;
                      r -= vscrinfo.yres;
                      *r = high;
                      r -= vscrinfo.yres;
                      *r = high;
                      r -= vscrinfo.yres;

                      if(y < varblock.min_i) varblock.min_i = y;
                      else if(y > varblock.max_i) varblock.max_i = y;
                      if(x > varblock.max_j) varblock.max_j = x;
                      else if(x < varblock.min_j) varblock.min_j = x;
                  } else {
                      r -= (vscrinfo.yres << 2);
                  }
                  f++, c++;
              }
              r_column++;
          }
          if(varblock.min_i < 9999) {
              if(varblock.max_i < 0) varblock.max_i = varblock.min_i;
              if(varblock.max_j < 0) varblock.max_j = varblock.min_j;
              rfbMarkRectAsModified(rfbScreen, (varblock.min_i * 2) + display_x_offset,
                  ((varblock.min_j - 2) * 2) + display_y_offset, (varblock.max_i * 2 + 1) +
                  display_x_offset + 1, (varblock.max_j * 2) + display_y_offset + 1);
              rfbProcessEvents(rfbScreen, 10000);
          }
      } else {
          unsigned int *fl, *fr, *cl, *cr;
          unsigned int *r_column, *r;
          int x, y;

          fl = (unsigned int *)fbbuf;
          fr = fl + vscrinfo.xres/2;
          cl = (unsigned int *)cmpfbbuf;
          cr = cl + vscrinfo.xres/2;
          r_column = (unsigned int *)(rfbbuf + (rfb_width) * ((vscrinfo.xres + display_y_offset) - 1) + display_x_offset);
          for(y = 0; y < vscrinfo.yres; y += 2) {
              r = r_column;
              for(x = vscrinfo.xres - 1; x >= 0; x -= 2) {
                  if((*fl != *cl) || (*fr != *cr)) {
                      /* One of the pixels needs updating.  Let's not waste
                       * time figuring out which of the two has changed;
                       * instead, to save cycles, let's just treat both of
                       * them as being changed. */
                      unsigned int double_pixel_left, double_pixel_right;
                      *cl   = *fl;
                      *cr   = *fr;
                      double_pixel_left  = *fr;
                      double_pixel_right = *fl;
                      *r =
                             ((double_pixel_right >> varblock.r_offset) & 0x00001f)
                          | (((double_pixel_right >> varblock.g_offset) & 0x00001f) <<  5)
                          | (((double_pixel_right >> varblock.b_offset) & 0x00001f) << 10)
                          |((((double_pixel_left  >> varblock.r_offset) & 0x00001f)
                          | (((double_pixel_left  >> varblock.g_offset) & 0x00001f) <<  5)
                          | (((double_pixel_left  >> varblock.b_offset) & 0x00001f) << 10)) << 16)
                          ;
                      r -= rfb_width / 2;
                      *r =
                             ((double_pixel_left  >> varblock.r_offset) & 0x1f0000)
                          | (((double_pixel_left  >> varblock.g_offset) & 0x1f0000) <<  5)
                          | (((double_pixel_left  >> varblock.b_offset) & 0x1f0000) << 10)
                          | ((((double_pixel_right >> varblock.r_offset) & 0x1f0000)
                          | (((double_pixel_right >> varblock.g_offset) & 0x1f0000) <<  5)
                          | (((double_pixel_right >> varblock.b_offset) & 0x1f0000) << 10)) >> 16)
                          ;
                      r -= rfb_width / 2;
                      if(y < varblock.min_i) varblock.min_i = y;
                      else if(y > varblock.max_i) varblock.max_i = y;
                      if(x > varblock.max_j) varblock.max_j = x;
                      else if(x < varblock.min_j) varblock.min_j = x;
                  } else {
                      r -= rfb_width;
                  }
                  fl++, cl++;
                  fr++, cr++;
              }
              fl += vscrinfo.xres/2;
              cl += vscrinfo.xres/2;
              fr += vscrinfo.xres/2;
              cr += vscrinfo.xres/2;
              r_column++;
          }

          /* Mark the changed area as modified */
          if(varblock.min_i < 9999) {
              if(varblock.max_i < 0) varblock.max_i = varblock.min_i;
              if(varblock.max_j < 0) varblock.max_j = varblock.min_j;
              varblock.max_j++;
              varblock.min_j--;
              varblock.max_i++;
              varblock.min_i--;
              rfbMarkRectAsModified(rfbScreen, varblock.min_i + display_x_offset,
                  varblock.min_j + display_y_offset, varblock.max_i +
                  display_x_offset + 1, varblock.max_j + display_y_offset + 1);
              rfbProcessEvents(rfbScreen, 10000);
          }
      }

#ifndef IPAQ
      /* If we do the fancy display, update the LEDs lights in the rfb */
      if(fancy_display)
      {
          if(!noleds) {
            led_states=get_led_states();
            for(j=0;j<2;j++)
            {
                short int k;
                if((led_states&(1<<j)) != led[j].prev_state)
                {
                    if((led[j].prev_state=(led_states&(1<<j))))
                        if(j==0)
                            k=LED_ORANGE_COLOR;
                        else
                            k=LED_GREEN_COLOR;
                    else
                        k=0x08 | (0x08<<5) | (0x08<<10);    /* Soft black */

                    for(i=0;led[j].pixels[i][0];i++)
                        rfbbuf[led[j].pixels[i][0]+led[j].pixels[i][1]*rfb_width]=k;

                    rfbMarkRectAsModified(rfbScreen,led[j].min_x,led[j].min_y,
                        led[j].max_x+1,led[j].max_y+1);
                    rfbProcessEvents(rfbScreen,10000);
                }
            }
          }
      }
#endif
  }
}
