/* Zaurus touchscreen simulation daemon. This
   daemon uses /dev/sharp_ts as a "natural"
   source of touchscreen events, and /dev/tssim
   as a source of "artificial" events. If the
   event is "natural", it simply sends it to
   /dev/ts (which is a named pipe). If the
   event is "artificial", the range of the
   X must be 0 -> 240 and the range of the
   Y must be 0 -> 320 : the daemon will re-range
   them with the auto-calibrator's values.

   The main purpose of this program is to be
   used in combination with fbvncserver, so the
   Zaurus can be remote-controlled from a VNC
   client.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version
   2 of the License, or (at your option) any later version.

   (c) Pierre-Philippe Coupard, 30/01/2002 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>



/* Harware-dependent touchscreen event types */
#ifdef IPAQ
typedef struct
{
  unsigned short pressure;
  unsigned short x;
  unsigned short y;
  unsigned short pad;
} TS_EVENT;
#define RAW_TS_EVENTS_SRC "/dev/touchscreen/0raw.REAL"
#define CAL_TS_EVENTS_SRC "/dev/touchscreen/0.REAL"
#define RAW_TS_EVENTS_OUT "/dev/touchscreen/0raw"
#define CAL_TS_EVENTS_OUT "/dev/touchscreen/0"
#define PRESSURE_THRESHOLD	1
#else	/* #ifdef ZAURUS */
typedef struct
{
  long x;
  long y;
  long pressure;
  long long millisecs;
} TS_EVENT;
#define RAW_TS_EVENTS_SRC "/dev/sharp_ts"
#define RAW_TS_EVENTS_OUT "/dev/ts"
#define PRESSURE_THRESHOLD	500
#endif

/* Artificial touchscreen event type */
typedef struct
{
  long x;
  long y;
  long pressure;
} ARTIFICIAL_TS_EVENT;



/* Variables */
int outtsfd_raw=-1;
#ifdef IPAQ
int outtsfd_cal=-1;
#endif

#if 0
static int read_pointercal(int *j, int *k, int *l, int *m)
{
    FILE *file;
    int a, b, c, d, e, f, s;

    file = fopen("/etc/pointercal", "r");
    if(!file) return 0;

    if(7 != fscanf(file, "%d %d %d %d %d %d %d", &a, &b, &c, &d, &e, &f, &s)) {
        printf("Bad Read\n");
        fclose(file);
        return 0;
    }
    fclose(file);

printf("%d %d %d %d %d %d %d\n", a, b, c, d, e, f, s);

    /*
    a = s * (screen_tl.x() - screen_br.x() ) / (dev_tl.x() - dev_br.x());
    b = 0;
    c = s * screen_tl.x() - a * dev_tl.x();

    d = 0;
    e = s * (screen_tl.y() - screen_br.y() ) / (dev_tl.y() - dev_br.y());
    f = s * screen_tl.y() - e * dev_tl.y();

    Solve for dev_tl.{xy}:

    dev_tl.x() = (s * screen_tl.x() - c) / a
    dev_tl.y() = (s * screen_tl.y() - f) / e

    */

    /* raw_min_x */
    *j = ((s * 0) - c) / a;
    /* raw_min_y */
    *k = ((s * 0) - f) / e;
    /* raw_max_x */
    *l = ((s * 320) - c) / a;
    /* raw_max_y */
    *m = ((s * 240) - f) / e;

    return 0;
}
#endif



/* Read 4 integers from a text file. */
int readrcfile(char *fname,int *a,int *b,int *c,int *d)
{
  FILE *rcfd;
  char buf[128];
  int i,j;

  /* Open the file */
  if((rcfd=fopen(fname,"r"))==NULL)
    return(-1);

  for(i=0;i<4;i++)
  {
    /* Read a line */
    if(fgets(buf,128,rcfd)==NULL)
    {
      fclose(rcfd);
      return(-2);
    }

    /* Zap trailing '\r's, '\n's and whitespaces */
    j=strlen(buf);
    while(j && isspace(buf[j-1]))
      buf[--j]=0;

    /* If we have an empty line, ignore it */
    if(j==0)
      i--;
    else
    {
      /* Check that we have only characters that are 0-9 (or - if it's the 1st*/
      for(j--;j>=0;j--)
        if(!(isdigit(buf[j]) || (buf[j]=='-' && j==0)))
        {
          fclose(rcfd);
          return(-2);
        }

      /* Turn the string into an integer */
      j=atoi(buf);
      switch(i)
      {
        case 0:*a=j;break;
        case 1:*b=j;break;
        case 2:*c=j;break;
        case 3:*d=j;break;
      }
    }
  }

  fclose(rcfd);
  return(0);
}



/* Write 4 integers in a text file */
int writercfile(char *fname,int a,int b,int c,int d)
{
  int fd;
  char buf[32];

  if((fd=open(fname,O_WRONLY | O_CREAT | O_TRUNC,S_IRUSR | S_IWUSR))<0)
    return(-1);

  /* Write the raw values in the file */
  sprintf(buf,"%d\n",a);
  write(fd,buf,strlen(buf));
  sprintf(buf,"%d\n",b);
  write(fd,buf,strlen(buf));
  sprintf(buf,"%d\n",c);
  write(fd,buf,strlen(buf));
  sprintf(buf,"%d\n",d);
  write(fd,buf,strlen(buf));
  close(fd);

  return(0);
}



/* Write to an events output fifo that may or may not have a listener. If
   cal!=0, the calibrated events output fifo is used, otherwise the raw events
   output fifo is used. If size==0, we still try to reopen the fifo if it's
   closed, but we don't write anything */
int write_to_fifo(int cal,char *buf,int size)
{
  int *fd;
  char *fname;

  fd=&outtsfd_raw;
  fname=RAW_TS_EVENTS_OUT;
#ifdef IPAQ
  if(cal)
  {
    fd=&outtsfd_cal;
    fname=CAL_TS_EVENTS_OUT;
  }
#endif

  /* Do we need to reopen the file first ? */
  if(*fd<0)
    if((*fd=open(fname,O_WRONLY | O_NONBLOCK))<0)
      return(-1);

  if(size!=0 && write(*fd,buf,size)==-1 && errno==EPIPE)
  {
    close(*fd);
    *fd=-1;
    return(-1);
  }

  return(0);
}



/* Main program */
int main(int argc,char **argv)
{
  int raw_max_x,raw_max_y,raw_min_x,raw_min_y;
  int cal_vals_ok=0;
  int save_cal_vals;
  int from_pointercal;
  int stsfd;
  int ztsfd_raw;
  TS_EVENT raw_tsev;
#ifdef IPAQ
  int ztsfd_cal;
  TS_EVENT cal_tsev;
#endif
  ARTIFICIAL_TS_EVENT art_tsev;
  fd_set fds;
  struct timeval fd_timeout;
  int max_fd;
  int daemonize=1;
  int i;

  /* See if we have an argument */
  for(i=1;i<argc;i++)
  {
    if(!strcmp(argv[i],"-h") || !strcmp(argv[i],"--help"))
    {
      printf("%s: error !\n",argv[0]);
      printf("Usage: %s [-f]",argv[0]);
      printf("\n");
      printf("  -f : stay in the foreground\n");
      printf("       (don't become a daemon)\n");
      return(-1);
    }
    else if(!strcmp(argv[i],"-f"))
      daemonize=0;
  }

  from_pointercal = 0;
#if 0
  if(read_pointercal(&raw_min_x, &raw_min_y, &raw_max_x, &raw_max_y)) {
printf("Got pointercal: %d, %d, %d, %d\n", raw_min_x, raw_min_y, raw_max_x, raw_max_y);
    cal_vals_ok = 1;
    from_pointercal = 1;
  } else {
#endif
    /* Read the calibration values from /etc/touch_cal.conf for the 1st time if
        we can. If we can't, we just assume the file doesn't exist yet */
    switch(readrcfile("/etc/tssim_cal.conf",&raw_min_x,&raw_min_y,
                                            &raw_max_x,&raw_max_y))
    {
    case 0:
      cal_vals_ok=1;
      break;
    case -2:
      fprintf(stderr,"Reading /etc/tssim_cal.conf: syntax error\n");
      return(-1);
      break;
    }
#if 0
  }
#endif

  /* Open the source of real raw touchscreen events */
  if((ztsfd_raw=open(RAW_TS_EVENTS_SRC,O_RDONLY))<0)
  {
    perror(RAW_TS_EVENTS_SRC);
    return(-1);
  }

#ifdef IPAQ
  /* Open the source of real calibrated touchscreen events */
  if((ztsfd_cal=open(CAL_TS_EVENTS_SRC,O_RDONLY))<0)
  {
    perror(CAL_TS_EVENTS_SRC);
    return(-1);
  }
#endif

  /* Open /dev/tssim as our source of artificial touchscreen events */
  if((stsfd=open("/dev/tssim",O_RDWR))<0)
  {
    perror("Opening /dev/tssim");
    return(-1);
  }

  /* Daemonize unless we've been told not to */
  if(daemonize)
  {
    switch(fork())
    {
    case -1:
      perror("fork");
      return(-1);
      break;

    case 0:
      setsid();
      chroot("/");
      chdir("/");
      umask(0);
      close(0);
      close(1);
      close(2);
      break;

    default:
      return(0);
      break;
    }
  }

  /* Ignore SIGPIPE, so our writes can return EPIPE */
  signal(SIGPIPE,SIG_IGN);

  /* Get touchscreen events from the real or artificial event sources and send
     then to the proper event output routine */
  if(stsfd>ztsfd_raw) max_fd=stsfd+1;
  else max_fd=ztsfd_raw+1;
#ifdef IPAQ
  if(ztsfd_cal>=max_fd) max_fd=ztsfd_cal+1;
#endif
  while(1)
  {
    /* Watch for frames from the event sources */
    FD_ZERO(&fds);
    FD_SET(stsfd,&fds);
    FD_SET(ztsfd_raw,&fds);
#ifdef IPAQ
    FD_SET(ztsfd_cal,&fds);
#endif
    fd_timeout.tv_sec=1;
    fd_timeout.tv_usec=0;
    switch(select(max_fd,&fds,NULL,NULL,&fd_timeout))
    {
    case -1:				/* select error */
      perror("select");
      sleep(1);				/* Just in case */
      break;

    case 0:				/* Timeout reached */
      /* Try to reopen the output pipe(s) to make sure nobody is blocking on
         an open at the other end (like the Xipaq server notably) */
      write_to_fifo(0,NULL,0);
#ifdef IPAQ
      write_to_fifo(1,NULL,0);
#endif
      break;

    default:				/* A touchscreen event has arrived */
      /* Is the frame coming from the real raw touchscreen device ? */
      if(FD_ISSET(ztsfd_raw,&fds))
      {
        /* Read the frame */
        if(read(ztsfd_raw,&raw_tsev,sizeof(raw_tsev))==sizeof(raw_tsev))
        {
          /* Send the frame as-is */
          write_to_fifo(0,(char *)&raw_tsev,sizeof(raw_tsev));

          /* If we don't have valid calibration values yet, or if the frame
             contains wider coordinates, change the calibration values */
          if(raw_tsev.pressure && !from_pointercal)
          {
            if(cal_vals_ok)
            {
              save_cal_vals=0;
              if(raw_min_x>raw_tsev.x)
              {
                raw_min_x=raw_tsev.x;
                save_cal_vals=1;
              }
              else if (raw_max_x<raw_tsev.x)
              {
                raw_max_x=raw_tsev.x;
                save_cal_vals=1;
              }
              if(raw_min_y>raw_tsev.y)
              {
                raw_min_y=raw_tsev.y;
                save_cal_vals=1;
              }
              else if (raw_max_y<raw_tsev.y)
              {
                raw_max_y=raw_tsev.y;
                save_cal_vals=1;
              }
              if(save_cal_vals)
                writercfile("/etc/tssim_cal.conf",raw_min_x,raw_min_y,
						raw_max_x,raw_max_y);
            }
            else
            {
              raw_min_x=raw_tsev.x;
              raw_min_y=raw_tsev.y;
              raw_max_x=raw_min_x+1;	/* +1 so we don't do div. by 0 later */
              raw_max_y=raw_min_y+1;	/* +1 so we don't do div. by 0 later */
              cal_vals_ok=1;
            }
          }
        }
      }

#ifdef IPAQ
      /* Is the frame coming from the real calibrated touchscreen device ? */
      if(FD_ISSET(ztsfd_cal,&fds))
      {
        /* Read the frame */
        if(read(ztsfd_cal,&cal_tsev,sizeof(cal_tsev))==sizeof(cal_tsev))
        {
          /* Send the frame as-is */
          write_to_fifo(1,(char *)&cal_tsev,sizeof(cal_tsev));
        }
      }
#endif

      /* Is the frame coming from the simulated touchscreen events fifo ? */
      if(FD_ISSET(stsfd,&fds))
      {
        /* Read the frame */
        if(read(stsfd,&art_tsev,sizeof(art_tsev))==sizeof(art_tsev))
        {
          /* Re-range the x and y to make a fake raw frame */
#ifdef IPAQ
          raw_tsev.y=(((240-art_tsev.x)*(raw_max_y-raw_min_y))/240)+raw_min_y;
          raw_tsev.x=((art_tsev.y*(raw_max_x-raw_min_x))/320)+raw_min_x;
          raw_tsev.pressure=art_tsev.pressure?PRESSURE_THRESHOLD:0;
          cal_tsev.x=320-art_tsev.y;
          cal_tsev.y=art_tsev.x;
          cal_tsev.pressure=art_tsev.pressure?PRESSURE_THRESHOLD:0;
#else	/* #ifdef ZAURUS */
printf("Artificial: %ld, %ld\n", art_tsev.x, art_tsev.y);
          raw_tsev.x=((art_tsev.x*(raw_max_x-raw_min_x))/240)+raw_min_x;
          raw_tsev.y=((art_tsev.y*(raw_max_y-raw_min_y))/320)+raw_min_y;
printf("Raw: %ld, %ld\n", raw_tsev.x, raw_tsev.y);
          raw_tsev.pressure=art_tsev.pressure?PRESSURE_THRESHOLD:0;
#endif

          /* Send the "raw"-ed frame */
          write_to_fifo(0,(char *)&raw_tsev,sizeof(raw_tsev));

#ifdef IPAQ
          /* Send the original artificial frame (calibrated by nature) */
          write_to_fifo(1,(char *)&cal_tsev,sizeof(cal_tsev));
#endif
        }
      }
      break;
    }
  }
}
